#!/usr/bin/env python
#-*- encoding: utf-8 -*-
"""
pypi_pkgbuild.py
Generates a PKGBUILD from a package distributed on the Cheeseshop.

It uses the simple view at http://pypi.python.org/simple/ to find the
needed package.

Copyright (c) 2008 Pierre Bourdon <delroth@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from distutils.core import DistutilsError
from pkg_resources import Distribution, Requirement
from setuptools import package_index, archive_util

import distutils.core
import os.path
import tarfile
import tempfile
import setuptools
import shutil
import string
import sys
import zipfile

def error(message):
    """
    Outputs an error on stderr.
    """
    print >>sys.stderr, 'error:', message

def monkey_patch():
    """
    Apply monkey patches at all the needed places.

    Most notably, it replaces distutils.setup and setuptools.setup in
    order to replace them by a PKGBUILD construction function, and it
    modifies Distribution.clone to keep the old package location (needed
    for the URL in the PKGBUILD).
    """

    distutils.core.setup = setuptools.setup = patched_setup
    Distribution.clone = patch_clone(Distribution.clone)

def patched_setup(*args, **kwargs):
    """
    A function replacing distutils.setup and setuptools.setup.

    When executed, there is a global _configuration dictionary, and our goal
    is to fill it with the informations passed to this setup function.
    """
    global _configuration
    for (k, v) in kwargs.iteritems():
        _configuration[k] = v

def patch_clone(f):
    """
    Decorator which will be applied to Distribution.clone in order to keep
    the old package location (PyPI URL).
    """
    def new(self, **kw):
        ret = f(self, **kw)
        ret.old_location = self.location
        return ret

    new.__name__ = f.__name__
    new.__doc__ = f.__doc__
    return new

def spec_from_requirement(req):
    """
    Returns a spec object infered from the requirement.
    """
    try:
        return Requirement.parse(req)
    except DistutilsError:
        raise ValueError("not a valid requirement (%r)" % req)

def dist_from_spec(spec, tmpdir):
    """
    Uses setuptools to get the distribution of the specified package (which
    could be on the Cheeseshop).

    If needed, this function downloads the package to the directory
    specified as `tmpdir'.
    """
    pi = package_index.PackageIndex()
    return pi.fetch_distribution(
        spec,
        tmpdir,
        force_scan=True,
        source=True,
        develop_ok=True
    )

def create_tmpdir(descr):
    """
    Creates a temporary directory described by the `descr' string.

    For example, on a Linux system with descr='foo', it will create a
    /tmp/foo-xxx directory.
    """
    return tempfile.mkdtemp(prefix=(descr + '-'))

def is_egg(zip):
    """
    Checks if a zip file is a Python egg.

    TODO: Implement a better detection.
    """
    return zip.endswith('.egg')

def tar_find_setupdir(path, tmpdir):
    """
    Returns the directory containing the setup.py script in a tar archive.
    """
    dir = None
    tar = tarfile.open(path)
    for member in tar:
        if member.isfile():
            name = member.name
            filename = os.path.basename(name)
            if filename == 'setup.py':
                dir = os.path.join(tmpdir, os.path.dirname(name))
                break
    tar.close()
    if dir is None:
        raise ValueError("unable to find a setup.py script in the archive")
    return dir

def zip_find_setupdir(path, tmpdir):
    """
    Returns the directory containing the setup.py script in a zip archive.
    """
    dir = None
    zip = zipfile.ZipFile(path)
    for member in zip.infolist():
        name = member.filename
        filename = os.path.basename(name)
        if filename == 'setup.py':
            dir = os.path.join(tmpdir, os.path.dirname(name))
            break
    zip.close()
    if dir is None:
        raise ValueError("unable to find a setup.py script in the archive")
    return dir

def egg_find_rootdir(path, tmpdir):
    """
    Returns the directory containing the EGG-INFO dir in a Python egg.
    """
    dir = None
    zip = zipfile.ZipFile(path)
    for member in zip.infolist():
        name = member.filename
        if os.path.basename(name) == '':
            name = os.path.dirname(name)
            dirname = os.path.basename(name)
            if name == 'EGG-INFO':
                dir = os.path.join(tmpdir, os.path.dirname(name))
                break
    zip.close()
    if dir is None:
        raise ValueError("unable to find an EGG-INFO dir in the egg")
    return dir

def extract(dist, tmpdir):
    """
    Extract the archive at `dist.location' to the temporary directory.

    Returns a couple of values, `path' which is the path to the extracted
    distribution, and `egg' which is True if the distribution is an egg.
    """
    try:
        archive_util.unpack_tarfile(dist.location, tmpdir)
        return tar_find_setupdir(dist.location, tmpdir), False
    except archive_util.UnrecognizedFormat:
        try:
            archive_util.unpack_zipfile(dist.location, tmpdir)
            if is_egg(dist.location):
                return egg_find_rootdir(dist.location, tmpdir). True
            else:
                return zip_find_setupdir(dist.location, tmpdir), False
        except archive_util.UnrecognizedFormat:
            raise ValueError("unrecognized archive format")

def find_package_metadata(dist, tmpdir):
    """
    Find the package metadata using a very simple behavior :
    - If the package is an egg, check the EGG-INFO directory
    - If the package is an archive, execute the setup file with a
    replaced setuptools.setup function.
    """
    path, egg = extract(dist, tmpdir)
    if egg:
        # NYI
        assert False
    else:
        oldcwd = os.getcwd()
        configuration = {'rootpath': path}
        try:
            os.chdir(path)
            g = globals()
            g['_configuration'] = configuration
            execfile('setup.py', g)
        finally:
            os.chdir(oldcwd)
        
        if configuration == {}:
            raise ValueError('the setup.py did not provide us any information')
        else:
            return configuration

PKGBUILD_TEMPLATE = '''\
# Maintainer: CHANGEME
# This PKGBUILD was auto-generated with `pypi-pkgbuild', report bugs here :
#   http://bitbucket.org/delroth/pypi-pkgbuild/

pkgname=%(name)s
pkgver=%(version)s
pkgrel=1
pkgdesc=%(description)r
arch=('i686' 'x86_64')
license=(%(license)r)
url=%(url)r
depends=('python' 'setuptools'%(dependencies)s)
source=%(download_url)r

build() {
    cd ${srcdir}/%(rootpath)s
    python setup.py install --root=${pkgdir} --optimize=1 || return 1
}
'''

def generate_pkgbuild(dist, cfg):
    """
    Generate a PKGBUILD file using the informations we got from the
    setup.py file or from the EGG-INFO directory, and the distribution
    parameters (containing, for example, the download URL).
    """

    def preprocess_name(name):
        """
        Sanitize a Python distribution name to become an Archlinux package.
        """
        trans_char = lambda c: (
            c if c in (string.letters + string.digits + '-') else '-'
        )
        return ''.join(map(trans_char, name.lower()))

    def generate_deps_string(dist, cfg):
        """
        Generate a string representing the package dependencies.
        """
        if 'install_requires' not in cfg or not cfg['install_requires']:
            return ''
        li = []
        for dep in cfg['install_requires']:
            req = Requirement.parse(dep)
            name = 'python-' + preprocess_name(req.project_name)
            if req.specs:
                (op, ver) = req.specs[0]
                if op == '==':
                    op = '='
                name += op + ver
            li.append("'%s'" % name)
        return ' ' + ' '.join(li)

    # Preprocess some parameters
    cfg['name'] = 'python-' + preprocess_name(cfg['name'])
    cfg['rootpath'] = os.path.basename(cfg['rootpath'])

    # Add others
    cfg['dependencies'] = generate_deps_string(dist, cfg)
    url = dist.old_location.split('#')[0]
    cfg['download_url'] = url

    pb = PKGBUILD_TEMPLATE % cfg

    # Output it somewhere
    os.mkdir(cfg['name'])
    path = os.path.join(cfg['name'], 'PKGBUILD')
    open(path, 'w').write(pb)
    print 'Generated a PKGBUILD for %r in %r' % (cfg['name'], path)

def usage():
    """
    Display a usage message to the user.
    """
    print """\
usage: pypi-pkgbuild [package1] [package2] ... [packageN]
Converts the specified packages to Archlinux PKGBUILD.

It does not currently handle Python eggs, only source packages. The
script fetches the packages automatically from the Python package index,
at http://pypi.python.org/.

See http://bitbucket.org/delroth/pypi-pkgbuild/ for a more complete
documentation.
"""

def main(argv):
    """
    Entry point for pypi-pkgbuild.

    Called by the executable script.
    """
    monkey_patch()
    tmpdir = create_tmpdir('pypi-pkgbuild')
    for requirement in argv:
        # Handle --help
        if requirement == '--help':
            usage()
            return 0

        try:
            spec = spec_from_requirement(requirement)
        except ValueError, err:
            error(err.message)
            return 1

        # XXX: setuptools send an AttributeError instead of failing
        # gracefully when it cannot find a distribution
        try:
            print 'Locating and fetching the %r distribution' % requirement
            dist = dist_from_spec(spec, tmpdir)
        except AttributeError:
            error('unable to find the %r distribution' % requirement)
            return 1

        # Get the metadata from the downloaded distribution
        print 'Extracting and fetching the metadatas'
        cfg = find_package_metadata(dist, tmpdir)

        # Generate the appropriate PKGBUILD file
        print 'Generating the Archlinux PKGBUILD'
        generate_pkgbuild(dist, cfg)
    # If everything went fine, delete the temporary directory
    shutil.rmtree(tmpdir)
    return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
